/* pdblib.c
   library for creating and compressing files according to the
   Palm Pilot PDB "standard" used by eg Aportis Doc and Rich Reader.
   Copyright (C) 1999-12-14 Dan Hovang (DocCompression)
   Copyright (C) 1999-12-14 Linus Walleij (file formats)

This file is part of the PDBlib package.
Latest version: http://www.df.lth.se/~triad/krad/
Mail us: triad@df.lth.se

PDBLIB is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

You should have received a copy of the GNU General Public License
along with PDBLIB; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA. 

*/

#include "config.h"
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <time.h>
#include "pilotpdb.h"

/* Static variables for this library */

/* Tells us whether the records in the list are compressed or not */
static int bIsCompressed = 0;
static char *cFileType = NULL;
static char *cFileID = NULL;
static DWORD RecordAttributes = 0x00000000UL;

/****************************************************************
 * This one returns the date in the "number of seconds sice
 * January 1st 1904" timestamp.
 ****************************************************************/

time_t PiTimeToUxTime (DWORD raw_time)
{
  if (raw_time>0x40000000)
    return (time_t)(raw_time - PILOT_TIME_DELTA);
  else
    return (time_t)raw_time;
}

DWORD UxTimeToPiTime (time_t t)
{
  return (unsigned long)((unsigned long)t + PILOT_TIME_DELTA);
}

static DWORD GetTimestamp()
{
  time_t LocalTime;
  
  time(&LocalTime);
  return UxTimeToPiTime(LocalTime);
}

/****************************************************************
 * This type defines the format of the linked list used for
 * storing the records in memory, then we have the manipulation
 * functions.
 ****************************************************************/

/* Create a new record in memory */
static tPilotRecord *NewRecord(byte *Chunk, DWORD ChunkSize, DWORD RecID)
{
  tPilotRecord *newp;

  newp = (tPilotRecord *) malloc(sizeof(tPilotRecord));
  newp->chunk = Chunk;
  newp->dwChunkSize = ChunkSize;
  if (RecID != 0)
    newp->dwRecordID = RecID;
  else
    newp->dwRecordID = RecordAttributes;
  newp->next = NULL;
  return newp;
}


/* Add a record to the list, create list of necessary */
tPilotRecord *AddRecord(tPilotRecord *listp, byte *inbuf, DWORD bufsize, DWORD RecID)
{
  tPilotRecord *p, *newp;
  byte *tmpbuf;

  /* Allocate space and copy the buffer area */
  tmpbuf = (byte *) malloc(bufsize);
  memcpy(tmpbuf, inbuf, bufsize);
  newp = NewRecord(tmpbuf, bufsize, RecID);

  /* Add the new record to the list */
  if(listp == NULL) {
    listp = newp;
    bIsCompressed = 0;
    return listp;
  }
  for (p=listp; p->next != NULL; p = p->next)
    ;
  p->next = newp;
  return listp;
}


/* Deallocate memory and destroy list */
void FreeRecords(tPilotRecord *listp)
{
  tPilotRecord *next;

  for (; listp != NULL; listp=next) {
    next = listp->next;
    free(listp->chunk);
    free(listp);
  }
}


static int ChunkListSize(tPilotRecord *listp)
{
  tPilotRecord *p;
  int nChunks;

  nChunks = 0;
  if(listp == NULL)
    return nChunks;
  nChunks++;
  for (p=listp; p->next != NULL; p = p->next)
    nChunks++;
  return nChunks;
}


static int AllChunksSize(tPilotRecord *listp)
{
  tPilotRecord *p;
  int nChunksSize;

  nChunksSize = 0;
  if(listp == NULL)
    return nChunksSize;
  for (p=listp; p->next != NULL; p = p->next)
    nChunksSize += (int) p->dwChunkSize;
  return nChunksSize;
}

/****************************************************************
 * The following routine yields a mint header to use.
 ****************************************************************/

static tPilotHeader MakeHeader(char *cTitle)
{
  tPilotHeader DocHeader;
  int i;

  strncpy(DocHeader.cName, cTitle, 31);
  /* Pad it with zeroes */
  for(i=strlen(DocHeader.cName);i<32;i++)
    DocHeader.cName[i] = '\0';
  DocHeader.cName[31] = '\0';
  DocHeader.wFlags = 0;
  DocHeader.wVersion = 0;
  DocHeader.dwCreateTime = GetTimestamp();
  DocHeader.dwModTime = GetTimestamp();
  DocHeader.dwBackupTime = 0;
  DocHeader.dwModNum = 0;
  DocHeader.dwAppInfo = 0;
  DocHeader.dwSortInfo = 0;
  for(i=0;i<4;i++) {
    DocHeader.cType[i] = cFileType[i];
    DocHeader.cID[i] = cFileID[i];
  }
  DocHeader.dwUniqueIDSeed = 0;
  DocHeader.dwNextRecordList = 0;
  DocHeader.wNumRecords = 0;
  return DocHeader;
}

/****************************************************************
 * The following routines writes and reads data to a pdb-file,
 * endianess-independent
 ****************************************************************/

/* Write a WORD big-endian */
static void PutBigWord(WORD x, FILE* OFile)
{
  fputc((x >> 8), OFile);
  fputc((x & 0xFF), OFile);
}


/* Write a DWORD, big-endian */
static void PutBigDWord(DWORD x, FILE* OFile)
{
  fputc((x >> 24 & 0xFF), OFile);
  fputc((x >> 16 & 0xFF), OFile);
  fputc((x >> 8 & 0xFF), OFile);
  fputc((x & 0xFF), OFile);
}


/* Read a WORD big-endian */
static WORD GetBigWord(FILE* IFile)
{
  register WORD w;

  w = (WORD) (fgetc(IFile) & 0xFF);
  w = ((WORD) (fgetc(IFile) & 0xFF)) | (w << 0x08);
  return(w);
}


/* Read a DWORD, big-endian */
static DWORD GetBigDWord(FILE* IFile)
{
  register DWORD dw;

  dw = (WORD) (fgetc(IFile) & 0xFF);
  dw = ((WORD) (fgetc(IFile) & 0xFF)) | (dw << 0x08);
  dw = ((WORD) (fgetc(IFile) & 0xFF)) | (dw << 0x08);
  dw = ((WORD) (fgetc(IFile) & 0xFF)) | (dw << 0x08);
  return(dw);
}


/* This one writes a header */
static void WriteHeader(tPilotHeader *Header, FILE *OutFile)
{
  register int i;

  /* Write header to target file */
  /* First shove out the name */
  fwrite(Header->cName,1,32,OutFile);
  PutBigWord(Header->wFlags, OutFile);
  PutBigWord(Header->wVersion, OutFile);
  PutBigDWord(Header->dwCreateTime, OutFile);
  PutBigDWord(Header->dwModTime, OutFile);
  PutBigDWord(Header->dwBackupTime, OutFile);
  PutBigDWord(Header->dwModNum, OutFile);
  PutBigDWord(Header->dwAppInfo, OutFile);
  PutBigDWord(Header->dwSortInfo, OutFile);
  /* Some text verbously */
  for(i=0;i<4;i++){
    fputc(Header->cType[i],OutFile);
  }
  for(i=0;i<4;i++){
    fputc(Header->cID[i],OutFile);
  }
  PutBigDWord(Header->dwUniqueIDSeed, OutFile);
  PutBigDWord(Header->dwNextRecordList, OutFile);
  PutBigWord(Header->wNumRecords, OutFile);
}


/* This one reads a header */
static tPilotHeader *ReadHeader(FILE *InFile)
{
  register int i;
  tPilotHeader *Header;

  Header = (tPilotHeader *) malloc(sizeof(tPilotHeader));
  fread(Header->cName,1,32,InFile);
  Header->wFlags = GetBigWord(InFile);
  Header->wVersion = GetBigWord(InFile);
  Header->dwCreateTime = GetBigDWord(InFile);
  Header->dwModTime = GetBigDWord(InFile);
  Header->dwBackupTime = GetBigDWord(InFile);
  Header->dwModNum = GetBigDWord(InFile);
  Header->dwAppInfo = GetBigDWord(InFile);
  Header->dwSortInfo = GetBigDWord(InFile);
  /* Some text verbously */
  for(i=0;i<4;i++){
    Header->cType[i] = fgetc(InFile);
  }
  for(i=0;i<4;i++){
    Header->cID[i] = fgetc(InFile);
  }
  /* Allocate memory for the Type and ID if not done before */
  if (cFileType == NULL)
    cFileType = malloc(5 * sizeof(char));
  if (cFileID == NULL)
    cFileID = malloc(5 * sizeof(char));
  for(i=0;i<4;i++){
    cFileType[i] = Header->cType[i];
    cFileID[i] = Header->cID[i];
  }
  /* Terminate the global strings */
  cFileType[4] = '\0';
  cFileID[4] = '\0';
  Header->dwUniqueIDSeed = GetBigDWord(InFile);
  Header->dwNextRecordList = GetBigDWord(InFile);
  Header->wNumRecords = GetBigWord(InFile);
  return(Header);
}


/* This one writes record 0 */
static void WriteRecord0(tPDBRecord0 *Record, FILE *OutFile)
{
  PutBigWord(Record->wVersion, OutFile);
  PutBigWord(Record->wSpare, OutFile);
  PutBigDWord(Record->dwPDBSize, OutFile);
  PutBigWord(Record->wNumRecs, OutFile);
  PutBigWord(Record->wRecSize, OutFile);
  PutBigDWord(Record->dwSpare2, OutFile);
}


/* This one reads record 0 */
static tPDBRecord0 *ReadRecord0(FILE *InFile, DWORD start)
{
  fpos_t OldFilePos;
  tPDBRecord0 *Record;

  fgetpos(InFile, &OldFilePos);
  fseek(InFile, start, SEEK_SET);
  Record = (tPDBRecord0 *) malloc(sizeof(tPDBRecord0));
  Record->wVersion = GetBigWord(InFile);
  Record->wSpare = GetBigWord(InFile);
  Record->dwPDBSize = GetBigDWord(InFile);
  Record->wNumRecs = GetBigWord(InFile);
  Record->wRecSize = GetBigWord(InFile);
  Record->dwSpare2 = GetBigDWord(InFile);
  fsetpos(InFile, &OldFilePos);
  return(Record);
}

void SetPDBFileType(char *cType, char *cID, DWORD attributes)
{
  /* Allocate memory for the Type and ID if not done before */
  if (cFileType == NULL)
    cFileType = malloc(5 * sizeof(char));
  if (cFileID == NULL)
    cFileID = malloc(5 * sizeof(char));
  strncpy(cFileType, cType, 4);
  cFileType[4] = '\0';
  strncpy(cFileID, cID, 4);
  cFileID[4] = '\0';
  RecordAttributes = attributes;
}


void WritePDBFile(tPilotRecord *listp, FILE *OutFile, char *FileDescription)
{
  DWORD BlockPointer;
  WORD nRecords;
  tPilotRecord *iterator;

  /* Header + First record of file */
  tPilotHeader PDBHeader;
  tPDBRecord0 PDBRecord0;

  /* Create a PDB header */
  PDBHeader = MakeHeader(FileDescription);

  /* Calculate the number of MAGIC_CHUNKSIZE byte records required to store this file */

  nRecords = ChunkListSize(listp);

  /* Increase by one to store record 0 on aportis docs */
  if (!strcmp(cFileType, "TEXt"))
    nRecords ++;

  PDBHeader.wNumRecords = nRecords;
  
  WriteHeader(&PDBHeader, OutFile);
  
  /*****************************************************************
   * BlockPointer is the block position relative to the beginning 
   * of the file. To start with we add up the size of the PILOT 
   * header, the record pointers, and the two padding bytes.
   *****************************************************************/
  
  BlockPointer = (unsigned long) PILOTHEADSZ + nRecords * 8UL + 2UL;
  
  /* Begin writing offsets for records */

  if (!strcmp(cFileType, "TEXt")) {
    /* Only on aportis docs a record0 appears */
    PutBigDWord(BlockPointer, OutFile);
    PutBigDWord(RecordAttributes, OutFile);
    /* Add the size of record 0 */
    BlockPointer += REC0SZ;
  }

  iterator = listp;

  /* Write out each and every block pointer and index */
  while(iterator != NULL) {
    PutBigDWord(BlockPointer, OutFile);
    PutBigDWord(iterator->dwRecordID, OutFile);
    BlockPointer += iterator->dwChunkSize;
    iterator = iterator->next;
  }

  /* Terminate with a zero word (the two padding bytes) */
  PutBigWord(0x0000U, OutFile);

  /* Write record 0 on aportis Docs */
  if (!strcmp(cFileType, "TEXt")) {
    /* Fill on some vital data in Record 0 */
    PDBRecord0.wVersion = bIsCompressed ? 2U : 1U;  /* 1=uncompressed 2=compressed*/
    PDBRecord0.wSpare = 0U;
    PDBRecord0.dwPDBSize = AllChunksSize(listp);
    PDBRecord0.wNumRecs = PDBHeader.wNumRecords - 1U;
    PDBRecord0.wRecSize = MAGIC_CHUNKSIZE;
    PDBRecord0.dwSpare2 = 0UL;
    WriteRecord0(&PDBRecord0, OutFile);
  }

  iterator = listp;

  /* Write out each and every record */
  while(iterator != NULL) {
    fwrite(iterator->chunk, sizeof(byte), iterator->dwChunkSize, OutFile);
    iterator = iterator->next;
  }
}

/* If called with end = 0 it reads to EOF */
static DWORD GetRecord(FILE *fin, byte *buf, DWORD start, DWORD end)
{
  fpos_t OldFilePos;
  register DWORD size;

  /* Save current file position */
  fgetpos(fin, &OldFilePos);
  if (end != 0) {
    fseek(fin, start, SEEK_SET);
    size = end-start;
    fread(buf, 1, size, fin);
  } else {
    /* Read one char at a time */
    fseek(fin, start, SEEK_SET);
    size = 0;
    while (!feof(fin)) {
      buf[size] = fgetc(fin);
      size++;
    }
  }
  /* Restore file position */
  fsetpos(fin, &OldFilePos);
  return size;
}

tPilotRecord *ReadPDBFile(FILE *InFile)
{
  register int i;
  DWORD RecID, RecSZ, BlockPointer, NextBlock;
  tPilotRecord *listp = NULL;
  tPilotHeader *PDBHeader;
  tPDBRecord0 *PDBRecord0;
  byte *inbuf;

  inbuf = (byte *) malloc(65536);
  PDBHeader = ReadHeader(InFile);
  
  fprintf(stderr, "Reading PDB: %s, Type: %s, ID: %s\n", PDBHeader->cName, cFileType, cFileID);

  /* Default to uncompressed data */
  bIsCompressed = 0;

  /* Read record pointers */
  for (i=0; i < PDBHeader->wNumRecords; i++) {
    BlockPointer = GetBigDWord(InFile);
    RecID = GetBigDWord(InFile);

    /* Aportis DOC and RichReader only */
    if (!strcmp(cFileType, "TEXt") && (i == 0)) {
      PDBRecord0 = ReadRecord0(InFile, BlockPointer);
      bIsCompressed = (PDBRecord0->wVersion >= 2); /* 1=uncompressed 2=compressed*/
      fprintf(stderr, "Doc chunksize: %d bytes (%s)\n", 
	     PDBRecord0->wRecSize, 
	     bIsCompressed ? "compressed" : "uncompressed");
    } else {
      if (i == PDBHeader->wNumRecords-1)
	NextBlock = 0;
      else {
	NextBlock = GetBigDWord(InFile);
	fseek(InFile, ftell(InFile)-sizeof(DWORD), SEEK_SET);
      }
      fprintf(stderr, "Getting record %d...", i);
      RecSZ = GetRecord(InFile, inbuf, BlockPointer, NextBlock);
      fprintf(stderr, " %ld bytes.\n", RecSZ);
      listp = AddRecord(listp, inbuf, RecSZ, RecID);
    }
  }

  fprintf(stderr, "Number of records: %d\n", ChunkListSize(listp));

  return(listp);
}

/****************************************************************
 * Dorky pack routines
 *
 * Allright the packing algorithm used by the Doc format is
 * indeed primitive. What can we do about it now? The packing
 * algorithm was reverse-engineered and reimplemented in C
 * by Skyflash of Oneway in a matter of hours.
 ****************************************************************/

#define DIST_BITS 11
#define COUNT_BITS 3
#define MAX_LEN ((1 << COUNT_BITS) + 2)
#define MAX_DIST ((1 << DIST_BITS) - 1)

/****************************************************************
 * This function compresses an inbuffer and presents the result
 * in an outbuffer, given the input size and returning the
 * output buffer size.
 ****************************************************************/

void DocCompress(byte *inbuf, int insize, byte *outbuf, int *outsize)
{
  int outptr = 0, len, best_len, code, run_len = 0, spc = 0;
  byte *prev, *best = 0, *curr = inbuf, *match, *run, *t;
  byte *last = (byte *)((int)inbuf + insize);

  __inline void emit(byte c)
	{
	  outbuf[outptr++] = c;
	}

  __inline void flush_run(void)
	{
	  if(run_len)
		emit(run_len);
	  while(run_len)
		emit(*run++), run_len--;
	}

  while(curr != last) {

	best_len = len = 0;
	t = prev = match = curr;
	while(prev-- != inbuf && ((int)(curr - prev) <= MAX_DIST)) {

	  if(*prev == *match) {
		t = prev++, match++, len++;
		while(*prev == *match && match != last && len < MAX_LEN)
		  prev++,	match++, len++;
		match = curr;
	  }

	  if(len) {
		if(len > 2 && len > best_len)
		  best_len = len, best = t;
		len = 0, prev = t;
	  }

	}
	
	if(best_len) {

	  if(spc)
		emit(0x20), spc = 0;	  

	  flush_run();

	  code = (((int)(curr - best)) << COUNT_BITS) | (best_len - 3);

	  emit((code >> 8) | 0x80);
	  emit(code & 0xff);

	  curr += best_len;

	} else {

	  if(*curr > 0x7f || (*curr > 0 && *curr < 9)) {

		if(spc)
		  emit(0x20), spc = 0;

		if(!run_len)
		  run = curr;

		if(++run_len == 8)
		  flush_run();

		curr++;

	  } else {

		flush_run();

		if(*curr == 0x20 && !spc)
		  spc = 1, curr++;
		else {

		  if(spc && (*curr >= 0x40 && *curr <= 0x7f))
			emit(*curr | 0x80), spc = 0;
		  else {
			if(spc)
			  emit(0x20), spc = 0;
			emit(*curr);
		  }

		  curr++;

		}

	  }

	}

  }

  if(spc)
	emit(0x20);

  *outsize = outptr;

}

/****************************************************************
 * This function reverses the compression above
 ****************************************************************/

void DeDocCompress(byte *inbuf, int insize, byte *outbuf, int *outsize,
			   int maxoutsize)
{
  int i = 0, c, j = 0;

  while(i < insize && j < maxoutsize) {

	c = inbuf[i++];

	if(c > 0 && c < 9)
	  while(c--) outbuf[j++] = inbuf[i++];
	else if(c < 0x80)
	  outbuf[j++] = c;
	else if(c >= 0xc0)
	  outbuf[j++] = 0x20, outbuf[j++] = c ^ 0x80;
	else {
	  int m, n;
	  c <<= 8;
	  c += inbuf[i++];
	  m = (c & 0x3fff) >> COUNT_BITS;
	  n = c & ((1 << COUNT_BITS) - 1);
	  n += 3;
	  while(n--) {
		outbuf[j] = outbuf[j - m];
		j++;
	  }
	}

  }

  *outsize = j;

}

void DocCompressList(tPilotRecord *listp)
{
  tPilotRecord *iterator;
  byte outbuf[MAGIC_CHUNKSIZE * 2];
  byte *tmpbuf;
  int outsz = 0;

  bIsCompressed = 1;
  iterator = listp;

  /* Pack each chunk */
  while(iterator != NULL) {
    DocCompress(iterator->chunk, (int) iterator->dwChunkSize, outbuf, &outsz);
    free(iterator->chunk);
    tmpbuf = (byte *) malloc(outsz);
    memcpy(tmpbuf, outbuf, outsz);
    iterator->chunk = tmpbuf;
    iterator->dwChunkSize = outsz;
    iterator = iterator->next;
  }
}

void DeDocCompressList(tPilotRecord *listp)
{
  tPilotRecord *iterator;
  byte outbuf[MAGIC_CHUNKSIZE];
  /* Hm, I wonder what happens if you input some bigger chunks :-) */
  byte *tmpbuf;
  int outsz = 0;

  bIsCompressed = 0;
  iterator = listp;

  /* Pack each chunk */
  while(iterator != NULL) {
    DeDocCompress(iterator->chunk, (int) iterator->dwChunkSize, outbuf, &outsz, MAGIC_CHUNKSIZE);
    free(iterator->chunk);
    tmpbuf = (byte *) malloc(outsz);
    memcpy(tmpbuf, outbuf, outsz);
    iterator->chunk = tmpbuf;
    iterator->dwChunkSize = outsz;
    iterator = iterator->next;
  }
}
