/* pilotpdb.h
   library for creating and compressing files according to the
   Palm Pilot PDB "standard" used by eg Aportis Doc and Rich Reader.
   Copyright (C) 1999-12-14 Dan Hovang (compression)
   Copyright (C) 1999-12-14 Linus Walleij (file formats)

This file is part of the PDBlib package.
Latest version: http://www.df.lth.se/~triad/krad/
Mail us: triad@df.lth.se

PDBlib is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

You should have received a copy of the GNU General Public License
along with Yelah xmltools; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA. 

*/

#ifndef PALMPACK_INCLUDED
#define PALMPACK_INCLUDED 1

#ifdef __cplusplus
extern "C" {
#endif

/* Time library is used in interface */
#include <time.h>

/*
  I don't know why this is 4096, someone at Palm Computing
  may have chosen it at random. It's a convenient chunk
  size though (hex $1000), and I don't know whether it'll 
  work with bigger chunks in al OS:es, sync:ers etc.
*/
#define MAGIC_CHUNKSIZE 4096U

/* Exact value of "Jan 1, 1970 0:00:00 GMT" - "Jan 1, 1904 0:00:00 GMT" */
#define PILOT_TIME_DELTA (unsigned)(2082844800)

/*
  These are record attributes used for OR:ing with the
  UniqueID of the record. Records may be secret, in use,
  dirty or "to be deleted at next hotsync".
 */
#define RECATT_SECRET  0x10000000
#define RECATT_INUSE   0x20000000
#define RECATT_DIRTY   0x40000000
#define RECATT_DELETE  0x80000000

/* Good typedefs when working with Palm Pilot stuff */

typedef unsigned char byte;
typedef unsigned long DWORD;
typedef unsigned short WORD;

/* Data types */

/****************************************************************
 * The PDB format is quite straight forward. It's
 * good to know that these routines imply architecture to be
 * Little-Endian (Motorola, SPARC, etc).
 *
 * The format looks like this:
 *
 *  PDB-Header         78 bytes (similar to PRC-header) 
 *  Record pointers    n * 8 bytes
 *  Padding            2 bytes (both zeroes)
 *  Record 0           16 bytes
 *  Records            n * 4096 bytes
 *  
 * The document header (aka the .PRC format) is documented by
 * Theodore Ts'o at:
 * http://web.mit.edu/tytso/www/pilot/prc-format.html
 * it is similar to the .pdb format used by this implementation
 ****************************************************************/

/* Used when writing to file. sizeof() bogus */
#define PILOTHEADSZ 78

typedef struct {
  char cName[32];          /* 32 characters NULL-terminated name  */
  WORD wFlags;             /* 0x0001 for executables,             */
                           /* 0x0000 in our case                  */
  WORD wVersion;           /* 0x0001 for executables,             */
                           /* 0x0000 in our case                  */
  DWORD dwCreateTime;      /* Time of creation                    */
  DWORD dwModTime;         /* Time of last modification           */
  DWORD dwBackupTime;      /* Time of last back-up                */
  DWORD dwModNum;          /* Must be 0x00000000 for executables  */
  DWORD dwAppInfo;         /* Must be 0x00000000 for executables  */
  DWORD dwSortInfo;        /* Must be 0x00000000 for executables  */
  char  cType[4];          /* Type ID 4 chars                     */
                           /* "appl" for applications etc         */
  char  cID[4];	           /* Creator ID 4 chars                  */
  DWORD dwUniqueIDSeed;	   /* Used for something                  */
  DWORD dwNextRecordList;  /* Used for something                  */
  WORD  wNumRecords;	   /* The number of resources in a PRC    */
                           /* in our case the number of records   */
} tPilotHeader;

/****************************************************************
 * This one record follows the list of all records, just before
 * the records themselves (see above)
 ****************************************************************/

#define REC0SZ 0x0010

typedef struct {
  WORD wVersion;           /* 1=plain text, 2=compressed */
  WORD wSpare;
  DWORD dwPDBSize;	   /* in bytes, when decompressed */
  WORD wNumRecs;           /* text records only; equals tPilotHeader.wNumRecs-1 */
  WORD wRecSize;	   /* usually 0x1000 */
  DWORD dwSpare2;
} tPDBRecord0;

/****************************************************************
 * This is a datatype for the linked list making up the in-memory
 * file body:
 *              +-------------+      +------+
 *     head --> | next        | ---> | next | .... ---> NULL
 *              +-------------+      +------+
 *              | *chunk      |      | -"-  |
 *              | dwRecordID  |      | -"-  |
 *              | dwChunkSize |      | -"-  |
 *              +-------------+      +------+
 *
 ****************************************************************/

typedef struct tPilotRecord tPilotRecord;
struct tPilotRecord {
  byte *chunk;
  DWORD dwRecordID;
  DWORD dwChunkSize;
  tPilotRecord *next;
};

/* Public functions */

time_t PiTimeToUxTime (DWORD raw_time);

DWORD UxTimeToPiTime (time_t t);

tPilotRecord *AddRecord(tPilotRecord *listp, byte *inbuf, DWORD bufsize, DWORD RecID);

void FreeRecords(tPilotRecord *listp);

void SetPDBFileType(char *cType, char *cID, DWORD attributes);

void WritePDBFile(tPilotRecord *listp, FILE *OutFile, char *FileDescription);

tPilotRecord *ReadPDBFile(FILE *InFile);

void DocCompress(byte *inbuf, int insize, byte *outbuf, int *outsize);

void DeDocCompress(byte *inbuf, int insize, byte *outbuf, int *outsize, int maxoutsize);

void DocCompressList(tPilotRecord *listp);

void DeDocCompressList(tPilotRecord *listp);

#ifdef __cplusplus
}
#endif

#endif

